#ifndef _GGI_DM_INT_H
#define _GGI_DM_INT_H

#include <ggi/internal/internal.h>

extern ggi_extid ggiDMID;

#define GGIDM_VISIBLE					0x0001000
#define GGIDM_INVALID					0x0002000

struct dm_ext;

typedef struct _ggidm_drawmap {
	ggi_visual_t	vis;
	struct dm_visual *parent;
	struct dm_ext *hook;
	
	struct _ggidm_drawmap *prev;
	struct _ggidm_drawmap *next;
	
	int		flags;
	ggi_coord	pos;
	ggi_coord	size;
} *ggidm_drawmap;

typedef struct dm_rect {
	struct dm_rect *prev;
	struct dm_rect *next;
	
	ggi_coord pos;
	ggi_coord size;
} dmrect;

typedef struct dm_visual {
	struct dm_ext	*parent;

	struct dm_visual *prev;
	struct dm_visual *next;
	
	ggi_visual_t	vis;
	
	int 		flags;
	ggi_coord	size;		/* virtual size of this visual */
	
	dmrect		*freerects;	/* Not allocated rectangles */
	ggidm_drawmap	drawmaps;	/* drawmaps acquired to this visual */
} dmvisual;

typedef struct dm_ext {
	dmvisual	*dmvis;		/* always the visible visual */
	int		usecount;
	
	int		flags;
	ggi_color	*colormap;
	int		cm_start;
	int		cm_len;
	
	ggidm_drawmap	drawmaps;	/* unacquired drawmaps */
	ggidm_drawmap	fchain;		/* visible drawmaps / flipping chain */
	
	ggi_coord	visible;	/* the graphics mode */
	ggi_graphtype	type;
	int		frames;

	/* Set the mode, create flipping chain, acquire drawmaps */
	int (*setmode)(ggi_visual_t vis, int sizex, int sizey, int frames,
		       ggi_graphtype type);
	/* Add an off-screen drawmap of the specified size */
	ggidm_drawmap (*adddrawmap)(ggi_visual_t vis, int sizex, int sizey,
				    int flags);
	/* Flip to the next visible drawmap */
	int (*flip)(ggi_visual_t vis);
	/* Close the specified drawmap */
	int (*closedrawmap)(ggidm_drawmap dm);
	/* (Un-)Acquire the specified drawmap */
	int (*acquire)(ggidm_drawmap dm, int acquire);
	
	/* Get a visual to access the drawmap */
	ggi_visual_t (*getvisual)(ggidm_drawmap dm);
	
	/* Old functions */
	ggifunc_setflags *old_setflags;
	ggifunc_flush *old_flush;
	ggifunc_setpalvec *old_setpalvec;
} dmext;


#define LIBGGI_DMEXT(vis) ((dmext *)LIBGGI_EXT(vis,ggiDMID))

/******** Some convinience functions/macros *******/
#define REMOVE_LIST(list,rc) \
do {									\
	if ((rc) == (rc)->next) *(list) = 0;				\
	else {								\
		(rc)->next->prev = (rc)->prev;				\
		(rc)->prev->next = (rc)->next;				\
		if ((rc) == *(list)) *(list) = (rc)->next;		\
	}								\
} while(0)

#define INSERT_LIST(list,rc) \
do {									\
	if (!*(list))							\
		*(list) = (rc)->next = (rc)->prev = rc;			\
	else {								\
		(rc)->next = *(list);					\
		(rc)->prev = (*(list))->prev;				\
		(rc)->next->prev = (rc);				\
		(rc)->prev->next = (rc);				\
	}								\
} while(0)

#define FREE_LIST(type,list) \
do {									\
	while(*(list)) {						\
		type __item = *(list);					\
		REMOVE_LIST(list,__item);				\
		free(__item);						\
	}								\
} while(0)

#include <ggi/dm.h>

#endif /* _GGI_DM_INT_H */
